/*
 * Copyright (c) 2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 ** velodyneReadTaskCodels.c
 **
 ** Codels called by execution task velodyneReadTask
 **
 **/

/* System headers */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

/* Genom headers */
#include <portLib.h>

/* Internal headers */
#include "server/velodyneHeader.h"
#include "velodyne3DImage.h"
#include "velodyneBuffer.h"

velodyneReadTask task;
VELODYNE_INTERFACE* device = NULL;
velodyne3DImage* threeDImage;
velodyne3DImageParams imageParams;
static POSTER_ID	sensorPosPosterId;

static STATUS
velodyneCreateBuffer (int numPackets)
{
  task.buffer.nbPackets = numPackets;
  task.buffer.read = 0;
  task.buffer.write = 0;
  task.buffer.packets = malloc (numPackets * sizeof (velodynePacket));
  if (task.buffer.packets == NULL) {
    fprintf (stderr, "Could not allocate buffer of size %d\n", numPackets);
    return ERROR;
  }
}

static STATUS
velodyneDestroyBuffer ()
{
  if (task.buffer.packets != NULL) {
    free (task.buffer.packets);
    task.buffer.packets = NULL;
  }
  task.buffer.nbPackets = 0;
  task.buffer.read = 0;
  task.buffer.write = 0;
  return OK;
}

/*------------------------------------------------------------------------
 *
 * velodyneReadTaskInit  --  Initialization codel (fIDS, ...)
 *
 * Description:
 *
 * Returns:    OK or ERROR
 */

STATUS
velodyneReadTaskInit(int *report)
{
  memset (&task, 0, sizeof (velodyneReadTask));
  task.buffer.packets = NULL;
  return OK;
}

/*------------------------------------------------------------------------
 *
 * velodyneReadTaskEnd  --  Termination codel
 *
 * Description:
 *
 * Returns:    OK or ERROR
 */

STATUS
velodyneReadTaskEnd(void)
{
  velodyneDestroyBuffer ();
  if (device != NULL) {
    velodyne_close(device);
    device = NULL;
  }
  return OK;
}

/*------------------------------------------------------------------------
 * Init
 *
 * Description: opens the device -- and get the pom poster address if needed
 *
 * Reports:      OK
 *              S_velodyne_INPUT_FILE_NOT_FOUND
 *              S_velodyne_BAD_INIT_PARAMS
 *              S_velodyne_CAN_NOT_OPEN_DEVICE
 *              S_velodyne_POM_POSTER_NOT_FOUND
 */

/* velodyneInitExec  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
velodyneInitExec(velodyneInitParams *initParams, int *report)
{
  POM_SENSOR_POS sensorPos;
  double yawMainToSensor;
  VELODYNE_MODEL model;

  if (SDI_F->initParams.usePom == VELODYNE_TRUE) {
    /* Look for the POM sensor position poster */
    if (velodynePOM_SENSOR_POSPosterFind (initParams->sensorPosPoster,
                                          &sensorPosPosterId)
        == ERROR) {
      fprintf (stderr, "velodyne: cannot find pom sensor position poster\n");
      *report = S_velodyne_POM_POSTER_NOT_FOUND;
      return ETHER;
    }
    /* Read the POM sensor position poster to retrieve the heading offset */
    if (velodynePOM_SENSOR_POSPosterRead (sensorPosPosterId, &sensorPos)
        == ERROR) {
      fprintf (stderr, "velodyne: unable to read pom sensor position poster\n");
      return ETHER;
    }
    yawMainToSensor = -sensorPos.sensorToMain.euler.yaw * 180.0 / M_PI;
  } else {
    yawMainToSensor = 0.0;
  }

  if (device != NULL) {
    velodyne_close(device);
    device = NULL;
  }

  /* Retrieves the model */
  switch (SDI_F->initParams.model) {
  case VELODYNE_HDL32:
    model = VELODYNE_32;
    break;
  case VELODYNE_HDL64:
    model = VELODYNE_64;
    break;
  default:
    *report = S_velodyne_BAD_INIT_PARAMS;
    return ETHER;
    break;
  }

  /* Opens the device */
  switch (SDI_F->initParams.mode) {
  case VELODYNE_FROM_FILE:
    device = velodyne_open_file (
        model,
        SDI_F->initParams.inputFileName,
        SDI_F->initParams.discardAngularSector,
        SDI_F->initParams.firstDiscardedAngle + yawMainToSensor,
        SDI_F->initParams.lastDiscardedAngle + yawMainToSensor);
    break;
  case VELODYNE_LIVE:
    device = velodyne_open_network (
        model,
        SDI_F->initParams.discardAngularSector,
        SDI_F->initParams.firstDiscardedAngle + yawMainToSensor,
        SDI_F->initParams.lastDiscardedAngle + yawMainToSensor);
    break;
  default:			/* This should actually never occur */
    *report = S_velodyne_BAD_INIT_PARAMS;
    return ETHER;
    break;
  }
  if (device == NULL) {
    fprintf(stderr, "velodyne: can not open the device\n");
    *report = S_velodyne_CAN_NOT_OPEN_DEVICE;
    return ETHER;
  }

  /* Sets the 3D image and updates the poster */
  switch (model) {
  case VELODYNE_32:
    threeDImage->height = VELODYNE_NUM_LASERS_32;
    threeDImage->maxScanWidth = VELODYNE_MAX_NUM_PACKETS_PER_REVOLUTION_32
                                * VELODYNE_NUM_SHOTS_OF_A_LASER_PER_PACKET_32;
    break;
  case VELODYNE_64:
    threeDImage->height = VELODYNE_NUM_LASERS_64;
    threeDImage->maxScanWidth = VELODYNE_MAX_NUM_PACKETS_PER_REVOLUTION_64
                                * VELODYNE_NUM_SHOTS_OF_A_LASER_PER_PACKET_64;
    break;
  }
  threeDImage->width = 0;
  if (velodyneThreeDImagevelodyne3DImagePosterWrite(
          VELODYNE_THREEDIMAGE_POSTER_ID, threeDImage)
      != OK) {
    *report = S_velodyne_CAN_NOT_UPDATE_3D_POSTER;
    return ETHER;
  }

  /* Sets the 3D image parameters */
  switch (model) {
  case VELODYNE_32:
    imageParams.model = VELODYNE_32;
    imageParams.numLinesPerPacket = VELODYNE_NUM_LASERS_32;
    imageParams.numColumnsPerPacket
        = VELODYNE_NUM_SHOTS_OF_A_LASER_PER_PACKET_32;
    break;
  case VELODYNE_64:
    imageParams.model = VELODYNE_64;
    imageParams.numLinesPerPacket = VELODYNE_NUM_LASERS_64;
    imageParams.numColumnsPerPacket
        = VELODYNE_NUM_SHOTS_OF_A_LASER_PER_PACKET_64;
    break;
  }
  if (SDI_F->initParams.enableOrganization == VELODYNE_TRUE) {
    imageParams.organizationEnabled = VELODYNE_TRUE;
    velodyne_get_organization_values(model,
                                     imageParams.lineOrder,
                                     imageParams.columnOffsets);
  } else {
    imageParams.organizationEnabled = VELODYNE_FALSE;
  }

  return ETHER;
}

/*
 * XXX Unlocked access to SDI_F->running, but we can guess it is atomic
 * XXX No way to return a useful error to system as the thread run in
 * background
 */
void*
velodyneReadMain (void *report)
{
  static POM_SENSOR_POS pos;
  static int i = 0;

  while (SDI_F->running != 0) {
	if (task.needOneShot) {
		pthread_mutex_lock(&task.mtx);
		pthread_cond_wait(&task.cond, &task.mtx);
	} else {
		pthread_mutex_lock(&task.mtx);
	}

    if (velodyne_read_packet(device, &(task.buffer.packets[task.buffer.write].packet))
	!= VELODYNE_PAYLOAD_SIZE) {
      fprintf (stderr, "velodyne: could not read packet\n");
      SDI_F->running = 0;
	  pthread_mutex_unlock(&task.mtx);
      return NULL;
    }
    if (SDI_F->initParams.usePom == VELODYNE_TRUE) {
		if (i++ % 50 == 0) {
			if (velodynePOM_SENSOR_POSPosterRead(sensorPosPosterId, &pos)== ERROR) 
			{
				fprintf(stderr, "velodyne: unable to read pom sensorPos poster\n");
				SDI_F->running = 0;
				pthread_mutex_unlock(&task.mtx);
				return NULL;
			}
      } 
	  task.buffer.packets[task.buffer.write].position = pos;
    }

	task.buffer.write = (task.buffer.write + 1) % task.buffer.nbPackets;
	if (task.buffer.write == task.buffer.read)
		task.buffer.read =  (task.buffer.read + 1) % task.buffer.nbPackets;
    pthread_mutex_unlock(&task.mtx);
  }

  return NULL;
}

/*------------------------------------------------------------------------
 * StartAcquisition
 *
 * Description: continously grab the data and fill the data buffer
 *
 * Reports:      OK
 *              S_velodyne_DEVICE_NOT_OPEN
 *              S_velodyne_BAD_PACKET_READING
 *              S_velodyne_ACQUISITION_ALREADY_STARTED
 */

/* velodyneStartAcquisitionStart  -  codel START of StartAcquisition
   Returns:  EXEC ETHER  */
ACTIVITY_EVENT
velodyneStartAcquisitionMain (int *report)
{
  if (SDI_F->running  != 0) {
    fprintf (stderr, "velodyne: acquistion already started %d\n", SDI_F->running);
    *report = S_velodyne_ACQUISITION_ALREADY_STARTED;
    return FAIL;
  }

  if (device == NULL) {
    fprintf(stderr, "velodyne: device not opened\n");
    *report = S_velodyne_DEVICE_NOT_OPEN;
    return FAIL;
  }

  /* Allocates the buffer */
  velodyneDestroyBuffer ();
  switch (device->model) {
  case VELODYNE_32:
    velodyneCreateBuffer (VELODYNE_MAX_NUM_PACKETS_PER_REVOLUTION_32);
    break;
  case VELODYNE_64:
    velodyneCreateBuffer (VELODYNE_MAX_NUM_PACKETS_PER_REVOLUTION_64);
    break;
  }

  if (pthread_mutex_init(&task.mtx, NULL )) {
	fprintf(stderr, "velodyne: can't init mutex\n");
    *report = S_velodyne_NOT_ENOUGH_RESSOURCE;
    return FAIL;
  } ;

  task.needOneShot = false;

  if (pthread_cond_init(&task.cond, NULL)) {
	fprintf(stderr, "velodyne: can't init condition variable\n");
	pthread_mutex_destroy(&task.mtx);
    *report = S_velodyne_NOT_ENOUGH_RESSOURCE;
    return FAIL;
  };

  SDI_F->running  = 1;
  if (pthread_create(&task.thr, NULL, velodyneReadMain, NULL)) {
	fprintf(stderr, "velodyne: can't init readMain thread\n");
	pthread_cond_destroy(&task.cond);
	pthread_mutex_destroy(&task.mtx);
    *report = S_velodyne_NOT_ENOUGH_RESSOURCE;
	SDI_F->running = 0;
    return FAIL;

  } ;

  return ETHER;
}
