/*
 * Copyright (c) 2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef _VELODYNE_STRUCT_H
#define _VELODYNE_STRUCT_H

/* Internal headers */
#include "velodyneClient.h"

/******************************************************************************/
/* Some general constants */

#define VELODYNE_MAX_STRLEN 128

/******************************************************************************/
/* Initialization parameters */

typedef enum velodyneMode {
  VELODYNE_LIVE,
  VELODYNE_FROM_FILE
} velodyneMode;

typedef enum velodyneModel {
  VELODYNE_HDL32,
  VELODYNE_HDL64
} velodyneModel;

typedef enum velodyneBoolean {
  VELODYNE_FALSE,
  VELODYNE_TRUE
} velodyneBoolean;

#define VELODYNE_DEFAULT_MODE                   VELODYNE_FROM_FILE
#define VELODYNE_DEFAULT_MODEL                  VELODYNE_HDL32
#define VELODYNE_DEFAULT_USE_POM                VELODYNE_FALSE
#define VELODYNE_DEFAULT_UNUSED                 0
#define VELODYNE_DEFAULT_SENSOR_POS_POSTER      "pomVelodyneFramePos"
#define VELODYNE_DEFAULT_INPUT_FILE_NAME        "/tmp/velodyne.pcap"
#define VELODYNE_DEFAULT_ENABLE_ORGANIZATION    VELODYNE_FALSE
#define VELODYNE_DEFAULT_DISCARD_ANGULAR_SECTOR VELODYNE_FALSE
#define VELODYNE_DEFAULT_FIRST_DISCARDED_ANGLE  0.0
#define VELODYNE_DEFAULT_LAST_DISCARDED_ANGLE   0.0

typedef struct velodyneInitParams {
  velodyneMode    mode;
  velodyneModel   model;
  velodyneBoolean usePom;
  int             unused;
  char            sensorPosPoster[VELODYNE_MAX_STRLEN];
  char            inputFileName[VELODYNE_MAX_STRLEN];
  velodyneBoolean enableOrganization;
  velodyneBoolean discardAngularSector;
  double          firstDiscardedAngle;
  double          lastDiscardedAngle;
} velodyneInitParams;

/******************************************************************************/
/* Target detection parameters
 *
 * trackingOn: states whether the detection algorithms is launched or
 * not
 * eraseTargetPoints: remove all the points whose (x,y) coordinates
 * are within the target footprint (centered on its COG, of size
 * "radius")
 * verbosity: states what will be printed during the detection process
 *
 * Parameters to select the points that will be considered as target points:
 *
 * maxDistance: Points further that maxDistance are not considered as
 * target points
 * zMin: points lower than zMin are not considered as target points
 * min|maxReflectance: only points whose reflectance is within these
 * bounds are considered as target points
 *
 * radius: distance used to detect the clusters with the Ransac algorithm
 *
 * Parameters to assess whether a detected cluster is a target or not:
 *
 * nbRequestedPoints: min number of points per square meter required
 * to consider a cluster as a target
 * [min|mac][Width|height]: clusters not fitting within these
 * threshods are not retained as target
 *
 * Note: all the distances are computed with a L1 norm (not Cartesian
 * distances) */

typedef enum velodyneVerbosity {
  VELODYNE_QUIET,
  VELODYNE_VERBOSE,
  VELODYNE_VERY_VERBOSE
} velodyneVerbosity;

/* The following default values for the target are fine for the "man with a
   reflective jacket" target */
#define VELODYNE_DEFAULT_MAX_TARGET_DISTANCE 20.0
#define VELODYNE_DEFAULT_TARGET_Z_MIN        -1.0
#define VELODYNE_DEFAULT_MIN_REFLECTANCE     220
#define VELODYNE_DEFAULT_MAX_REFLECTANCE     255
#define VELODYNE_DEFAULT_TARGET_RADIUS       0.5
#define VELODYNE_DEFAULT_NB_REQUESTED_POINTS 2600
#define VELODYNE_DEFAULT_TARGET_MIN_WIDTH    0.3
#define VELODYNE_DEFAULT_TARGET_MAX_WIDTH    0.8
#define VELODYNE_DEFAULT_TARGET_MIN_HEIGHT   0.25
#define VELODYNE_DEFAULT_TARGET_MAX_HEIGHT   1.0

typedef struct velodyneTargetParams {
  float maxDistance;
  float zMin;
  int   minReflectance;
  int   maxReflectance;
  float radius;
  int   nbRequestedPoints;
  float minWidth;
  float maxWidth;
  float minHeight;
  float maxHeight;
} velodyneTargetParams;

#define VELODYNE_DEFAULT_DETECTION_MODE      VELODYNE_FALSE
#define VELODYNE_DEFAULT_TRACKING_MODE       VELODYNE_FALSE
#define VELODYNE_DEFAULT_ERASE_TARGET_POINTS VELODYNE_FALSE
#define VELODYNE_DEFAULT_VERBOSITY           VELODYNE_QUIET

typedef struct velodyneTargetDetectionParams {
  velodyneBoolean   detectionOn;
  velodyneBoolean   trackingOn;
  velodyneBoolean   eraseTargetPoints;
  velodyneVerbosity verbosity;
} velodyneTargetDetectionParams;

typedef enum velodyneTargetStatus {
  VELODYNE_TARGET_NOT_DETECTED,
  VELODYNE_TARGET_NOT_ENOUGH_POINTS,
  VELODYNE_TARGET_NOT_WIDE_ENOUGH,
  VELODYNE_TARGET_TOO_WIDE,
  VELODYNE_TARGET_NOT_HIGH_ENOUGH,
  VELODYNE_TARGET_TOO_HIGH,
  VELODYNE_TARGET_DETECTED
} velodyneTargetStatus;

#define VELODYNE_MAX_TARGET_POINTS 4000

/* Internal data used to track a reflective target */
typedef struct velodyneTarget {
  velodyneTargetStatus status;
  int                  detectedTargetPoints;
  float                meanReflectance;
  double               radius;
  velodyne3DPoint      targetPoints[VELODYNE_MAX_TARGET_POINTS];
  velodyne3DPoint      max;
  velodyne3DPoint      min;
  velodyne3DPoint      cog;
  velodyne3DPoint      stdDev;
  /* These 3 points are used to store the min and max coordinates of the point
     set of the target, its center of gravity (i.e. mean) and standard deviation
     -- it might be a bit awkward to use the velodyne3DPoint structure for these
     purposes, but it is a simple way to do. */
} velodyneTarget;

/******************************************************************************/
/* Save parameters */
#define VELODYNE_DEFAULT_BASE_NAME "/tmp/velodyneDump"
#define VELODYNE_DEFAULT_NUM       0

typedef struct velodyneSaveParams {
  char baseName[VELODYNE_MAX_STRLEN];
  int  num;
} velodyneSaveParams;

/******************************************************************************/
/* Parameters that define the angular size of a scan IN THE ROBOT FRAME. They
 * are in degrees, because velodyne returns angles in degrees (in hundreth of
 * degrees actually), hence there is no need to deal with degrees/radian
 * conversions. Note also that the two angles are prefixed first/last (and not
 * min/max): these angles define a sector that goes counter-clockwise from
 * firstAngle to lastAngle. 
 * Examples:
 * (-40;40) defines a 80 degrees sector centered on the robot X (forward) axis
 * (-150;150) defines a 150 degrees sector centered on the robot X (forward) axis
 * (-120; -60) defines a 60 degrees sector centered on the robot -Y (rightwards) axis
 * (40;-40) defines a 280 degrees sector centered on the robot -X (backward) axis
 * (150;-150) defines a 60 degrees sector centered on the robot -X (backward) axis
 * (-60; -120) defines a 300 degrees sector centered on the robot Y (leftwards) axis
 * Finally (-180,180) (or the reverse) defines a full 360 degrees scan */

#define VELODYNE_FIRST_ANGLE -180
#define VELODYNE_LAST_ANGLE  180

typedef struct velodyneOneShotParams {
  double firstAngle;
  double lastAngle;
} velodyneOneShotParams;

#endif /* _VELODYNE_STRUCT_H */
